<!DOCTYPE html>
<html>
  <head>
    <meta charset="UTF-8">
    <style>
      .highlight {
        background-color: #ff0;
      }
    </style>
    <title>{% trans %}Taguette Codebook{% endtrans %}</title>
  </head>
  <body>
    <h1>{% trans %}Taguette Codebook{% endtrans %}</h1>
{% for tag in tags %}
      <h2>{{ tag.path }}</h2>
      <p class="number">{{ ngettext("%(number)d highlight", "%(number)d highlights", tag.highlights_count, number=tag.highlights_count) }}</p>
      <p>{{ tag.description }}</p>
{% endfor %}
  </body>
</html>
