{% extends "base.html" %}

{% block title %}Delete project | Taguette{% endblock %}

{% block contents %}
<h1>{% trans %}Delete project: confirm?{% endtrans %}</h1>
{% set project_name %}<span style="font-family: monospace, monospace;">{{ project.name }}</span>{% endset %}
<p class="lead">{# TRANSLATORS: Name of the project to be deleted #}{% trans %}Name: {{ project_name }}{% endtrans %}</p>
<p>{% trans %}Are you sure you want to delete this project? It currently has <strong>{{ documents }} documents and {{ highlights }} highlights in {{ tags }} tags</strong>. There is no way to undo.{% endtrans %}</p>
<form action="{{ reverse_url('delete_project', project.id) }}" method="post">
  {{ xsrf_form_html() }}
  <div class="form-group ">
    <button type="submit" class="btn btn-danger">{% trans %}Delete forever{% endtrans %}</button>
    <a href="{{ reverse_url('project', project.id) }}" class="btn btn-secondary">{% trans %}Go back{% endtrans %}</a>
  </div>
</form>
{% endblock %}
