/* Add Taguette font */
@font-face {
  font-family: BadScript;
  font-display: swap;
  src: url(../webfonts/BadScript.ttf);
}

/* Add body font */
@font-face {
  font-family: Montserrat;
  font-display: fallback;
  src: url(../webfonts/Montserrat-Light.ttf);
}

/* Change link colors */
.container-body a:not(.btn) {
  color: #3498DB;
}

/* Disabled links */
#document-contents a:not([href]):not(.highlight) {
  color: #216490;
}

/* Change body font */
body {
  font-family: Montserrat !important;
  font-size:1.11em !important;
}

/* Some space before content */
.container-body {
  padding-top: 1em;
}

/* Border for the tab container */
.tab-content {
  border-style: solid;
  border-width: 1px;
  border-color: #fff #dee2e6 #dee2e6 #dee2e6;
  padding: 0.4em;
}

/* Footer */
.footer {
  font-size: 0.82em;
  text-align: center;
}

/* Navigation */

.navbar.navbar-expand-lg.navbar-dark.bg-primary {
  background-color: #125368 !important;
}

.navbar-dark .navbar-nav .nav-link {
  color: #fff !important;
}

.navbar-dark .navbar-nav .nav-link.disabled {
  color: #edbd9d !important;
}

.navbar-dark .navbar-nav .nav-link.active {
  color: #85c1e9 !important;
}

.navbar-dark .navbar-nav .nav-link.disabled:hover, .navbar-dark .navbar-nav .nav-link.disabled:focus {
  color: #e9ad85 !important;
}

.navbar-dark .navbar-nav .nav-link:hover, .navbar-dark .navbar-nav .nav-link:focus {
  color: #85c1e9 !important;
}

.navbar-dark .navbar-nav .dropdown-item.active {
  color: #fff !important;
  background-color: #6c757d !important;
}

/* Make icon lighter */
.navbar-dark .navbar-nav .nav-link .fas {
  opacity: 0.7;
}

/* Tab navs for project*/

.nav-link {
  padding: 0.5rem 0.5rem !important;
}

/* Colors for button primary */

.btn.btn-primary {
  background-color: #125368 !important;
}

.btn.btn-outline-primary {
  color: #125368;
  background-color: transparent;
  background-image: none;
  border-color: #125368;
}

.btn.btn-outline-primary:hover {
  color: #fff !important;
  background-color: #125368 !important;
  border-color: #125368 !important;
}

.btn.btn-outline-primary:focus, .btn-outline-primary.focus {
  box-shadow: none !important;
}

.btn.btn-outline-primary:not(:disabled):not(.disabled).active:focus, .btn-outline-primary:not(:disabled):not(.disabled):active:focus, .show > .btn-outline-primary.dropdown-toggle:focus {
  box-shadow: none !important;
}

.btn.btn-outline-secondary:not(:disabled):not(.disabled).active:focus, .btn-outline-secondary:not(:disabled):not(.disabled):active:focus, .show > .btn-outline-secondary.dropdown-toggle:focus {
  box-shadow: none !important;
}

.btn.btn-outline-primary:not(:disabled):not(.disabled).active:focus, .btn-outline-primary:not(:disabled):not(.disabled):active:focus, .show > .btn-outline-primary.dropdown-toggle:focus {
  color: #fff !important;
  background-color: #125368 !important;
  border-color: #125368 !important;
}

.btn.btn-outline-primary.disabled, .btn-outline-primary:disabled {
  color: #125368 !important;
  background-color: transparent !important;
}

.btn.btn-outline-primary:not(:disabled):not(.disabled):active, .btn-outline-primary:not(:disabled):not(.disabled).active,
.show > .btn-outline-primary.dropdown-toggle {
  color: #fff !important;
  background-color: #125368 !important;
  border-color: #125368 !important;
}

.btn.btn-outline-primary:not(:disabled):not(.disabled):active:focus, .btn-outline-primary:not(:disabled):not(.disabled).active:focus,
.show > .btn-outline-primary.dropdown-toggle:focus {
  box-shadow: none !important;
}

.dropdown-item:focus, .dropdown-item:hover {
  color: #fff !important;
  background-color: #125368 !important;
  border-color: #125368 !important;
}

/* Colors for button info */
.btn.btn-info {
  background-color: #1d87a9 !important;
}

/* Colors for badges */
.badge-secondary {
  color: #fff !important;
}

.badge-info {
  color: #fff !important;
}

.badge-dark {
  color: #fff !important;
}

.badge-warning {
  color: #fff !important;
}

.badge-danger {
  color: #fff !important;
}

.badge-secondary {
  color: #fff !important;
  background-color: #3498DB !important;
}

.badge-primary {
  color: #fff !important;
  background-color: #125368 !important;
}

.badge-primary[href]:hover, .badge-primary[href]:focus {
  color: #fff !important;
  text-decoration: none !important;
  background-color: #1a252f !important;
}

.badge-dark[href]:focus, .badge-dark[href]:hover {
  color: #abacaf !important;
  text-decoration: none !important;
}

/* Don't show the resize control on textareas */
textarea.form-control {
  resize: none;
}

/* Wrap long tag names and long doc names */
.tag-name {
  word-break: break-all;
}

a.document-link-a {
  word-break: break-all;
}

/* Work-around to make a whole element clickable while it contains links */
.overlay-outer {
  position: relative;
}

.overlay-bottom {
  position: absolute;
  left: 0; top: 0; bottom: 0; right: 0;
}

.overlay-top {
  position: relative;
  pointer-events: none;
  z-index: 1;
}

.overlay-top a {
  pointer-events: all;
}

li.document-link-current {
  background-color: #e7f2ff;
}

li.document-link-current a.document-link-a {
  font-weight: bold;
}

.document-view {
  border: 2px solid #dee2e6;
  padding: 2.2em 0.3em 1.2em 0.3em;
  position: relative;
}

li.tag-current {
  background-color: #e7f2ff;
}

.backlight {
  color: #bbb;
}

#export-button {
  position: absolute;
  top: .1em;
  right: .5em;
}

#export-button .dropdown-toggle {
  padding: 0.125rem 0.25rem;
}

a.dropdown-item:hover {
  color: white !important;
  text-decoration: none;
}

#document-view .highlight {
  background-color: #ff0;
  color: black;
}

#hlinfo {
  display: none;
  position: absolute;
  background-color: white;
  border: 1px solid grey;
  padding: 0.2em;
}

.list-group-item-primary a {
  color: #404041;
}

.list-group .sublist {
  padding-left: 0;
  margin-top: 5px;
  margin-right: -20px;
  margin-left: 5px;
  margin-bottom: -12px;
  border-top: 1px solid #dee2e6;
  display: flex;
  flex-direction: column;
}

.list-group .sublist li {
  padding: 12px 20px;
}

.expand-marker {
  width: 7px;
}

.highlight-entry {
  border: 1px solid rgba(0,0,0,.125);
  margin: 1rem 0;
  padding: 1rem 0.6rem 1rem 0.6rem;
}

/* Spinner animation */
#spinner {
  border: 3vh solid #dcdcdc;
  border-top: 3vh solid #125368;
  border-radius: 50%;
  width: 30vh;
  height: 30vh;
  animation: spinneranim 2s linear infinite;
}

@keyframes spinneranim {
  0% { transform: rotate(0deg); }
  100% { transform: rotate(360deg); }
}

#spinner-modal .modal-dialog {
  display: table;
  position: relative;
  margin: 0 auto;
  top: calc(50% - 15vh);
}

#spinner-modal .modal-dialog .modal-content{
  background-color: transparent;
  border: none;
  width: 30vh;
}

/* --------------------SCROLLBAR FOR MODALS--------------------------*/

.modal-body {
  max-height: calc(100vh - 210px);
  overflow-y: auto;
}

/* --------------------RESPONSIVE DISPLAY HEADINGS--------------------------*/
@media (max-width: 575.98px) {
  .display-1 {
  font-size: 3rem;
  font-weight: 300;
  line-height: 1.0;
  }
  .display-2 {
    font-size: 2.75rem;
    font-weight: 300;
    line-height: 1.0;
  }
  .display-3 {
    font-size: 2.25rem;
    font-weight: 300;
    line-height: 1.0;
  }
  .display-4 {
    font-size: 1.75rem;
    font-weight: 300;
    line-height: 1.0;
  }
}
