<!DOCTYPE html>
<html>
  <head>
    <meta charset="UTF-8">
    {%- if path %}
    <title>{% trans %}Taguette highlights: {{ path }}{% endtrans %}</title>
    {%- else %}
    <title>{% trans %}Taguette highlights{% endtrans %}</title>
    {% endif %}
    <style>
      h1 {
        margin-bottom: 1em;
      }
    </style>
  </head>
  <body>
    {%- if path %}
    <h1>{% trans %}Taguette highlights: {{ path }}{% endtrans %}</h1>
    {%- else %}
    <h1>{% trans %}Taguette highlights{% endtrans %}</h1>
    {% endif %}
{% for id, snippet, document, tags in highlights %}
    {{ snippet |safe }}
    <p>
      {% trans doc=document %}<strong>Document:</strong> {{ doc }}{% endtrans %}
      <strong>{% trans %}Tags:{% endtrans %}</strong>
      {%- for tag in tags %}
        {{ tag }}
        {%- if not loop.last %},{% endif %}
      {%- endfor %}
    </p>
    {%- if not loop.last %}
    <hr>
    {%- endif %}
{% endfor %}
  </body>
</html>
