<p>{% trans %}Someone has requested to reset the password on the Taguette account associated with this email. If this was you, please follow this link: <a href="{{ link }}">{{ link }}</a>{% endtrans %}</p>

<p>{% trans %}Your user name is: <strong>{{ login }}</strong>{% endtrans %}</p>

<p>{% trans %}This link will expire in 1 day.{% endtrans %}</p>
