<!DOCTYPE html>
<html>
  <head>
    <meta charset="UTF-8">
    <style>
      .highlight {
        background-color: #ff0;
      }
      .taglist {
        font-style: oblique;
        background: #fbb !important;
      }
    </style>
    <title>{{ name }}</title>
  </head>
  <body>
    <h1>{{ name }}</h1>
{{ contents }}
  </body>
</html>
