{% extends "base.html" %}

{% block title %}Reset password | Taguette{% endblock %}

{% block contents %}
{% if error %}
<div class="alert alert-danger">{{ error }}</div>
{% endif %}

<p>{% trans %}Please enter a new password below:{% endtrans %}</p>
<form action="{{ reverse_url('new_password') }}" method="post">
  {{ xsrf_form_html() }}
  <input type="hidden" name="reset_token" value="{{ reset_token }}" />
  <div class="form-group row">
    <label for="password1" class="col-sm-2 col-form-label">{% trans %}Password{% endtrans %}</label>
    <div class="col-sm-8">
      <input type="password" class="form-control" id="password1" name="password1" required />
    </div>
  </div>
  <div class="form-group row">
    <label for="password2" class="col-sm-2 col-form-label">{% trans %}Confirm password{% endtrans %}</label>
    <div class="col-sm-8">
      <input type="password" class="form-control" id="password2" name="password2" required />
    </div>
  </div>
  <div class="form-group row">
    <div class="col-sm-10 offset-sm-2">
      <button type="submit" class="btn btn-primary">{# TRANSLATORS: Submit button #}{% trans %}Set new password{% endtrans %}</button>
    </div>
  </div>
</form>
{% endblock %}
