{% extends "base.html" %}

{% block title %}Cookie warning | Taguette{% endblock %}

{% block contents %}
<h1>{% trans %}Cookie warning{% endtrans %}</h1>
<p>{% trans %}This website uses cookies to store your identity (if you choose to log in) and as a protection against <a href="https://en.wikipedia.org/wiki/Cross-site_request_forgery">cross-site request forgery (CSRF)</a>. Before continuing, you must accept their use by clicking the button below. Those <strong>are not</strong> "third-party" or "tracking" cookies.{% endtrans %}</p>

<form action="{{ reverse_url('cookies_prompt') }}" method="post">
  <input type="hidden" name="accepted" value="yes" />
  <input type="hidden" name="next" value="{{ next }}" />
  <button type="submit" class="btn btn-primary">{% trans %}Accept cookies{% endtrans %}</button>
</form>
{% endblock %}
