{% extends "base.html" %}

{% block title %}{{ error_title }} | Taguette{% endblock %}

{% block contents %}
<h1>{{ error_title }}</h1>
<p>{{ error_message }} Use the back button of your browser or <a href="/">click here to return to the index</a>.</p>
{% endblock %}
