{% extends "base.html" %}

{% block title %}Reset password | Taguette{% endblock %}

{% block contents %}
{% if message %}
<div class="alert alert-secondary">{{ message }}</div>
{% endif %}
{% if error %}
<div class="alert alert-danger">{{ error }}</div>
{% endif %}

<p>{% trans %}If you had set an email address in your profile, you can enter it below and a link to reset your password will be sent to you.{% endtrans %}</p>
<form action="{{ reverse_url('reset_password') }}" method="post">
  {{ xsrf_form_html() }}
  <div class="form-group row">
    <label for="email" class="col-sm-2 col-form-label">{# TRANSLATORS: Email address #}{% trans %}Email{% endtrans %}</label>
    <div class="col-sm-8">
      <input type="text" class="form-control" id="email" name="email" required />
    </div>
    <div class="col-sm-2">
      <button type="submit" class="btn btn-primary">{# TRANSLATORS: Submit button #}{% trans %}Send{% endtrans %}</button>
    </div>
  </div>
</form>
<p>{% trans %}If you didn't set an email address, you can try contacting your server administrator.{% endtrans %}</p>
{% endblock %}
